/*
 *	mvm.c
 */

#include <stdio.h>
#include <string.h>
#include "gen.h"

static int Step;
static int Disassemble;

#define MVM_MEMLEN	(1<<20)
#define MVM_NREG	(16)

static Word	Mem[MVM_MEMLEN/sizeof(Word)];
static Word	Reg[MVM_NREG];
static int	Z, N, C, V;

#define PC	(Reg[MVM_PC])
#define FP	(Reg[MVM_FP])
#define SP	(Reg[MVM_SP])

static char *Inst(Word inst)
{
    static char buf[1024];
    Address d20, k16;
    int ri, rk, rj;

    d20 = (inst & 0xfffff);
    if (inst & 0x80000)
	d20 |= ~0xfffff;
    ri = (inst >> 24) & 0xf;
    rj = (inst >> 20) & 0xf;

    switch ((inst >> 24) & 0xf0) {
    case MVM_LD:
	sprintf(buf, "LD R%d,R%d,%ld", ri, rj, d20);
	return buf;
    case MVM_ST:
	sprintf(buf, "ST R%d,R%d,%ld", ri, rj, d20);
	return buf;
    case MVM_LDB:
	sprintf(buf, "LDB R%d,R%d,%ld", ri, rj, d20);
	return buf;
    case MVM_LDUB:
	sprintf(buf, "LDUB R%d,R%d,%ld", ri, rj, d20);
	return buf;
    case MVM_STB:
	sprintf(buf, "STB R%d,R%d,%ld", ri, rj, d20);
	return buf;
    }

    k16 = (inst & 0xffff);
    ri = (inst >> 20) & 0xf;
    rj = (inst >> 16) & 0xf;
    rk = k16 & 0xf;

    switch ((inst >> 24) & 0xff) {
    case MVM_LDI:
	sprintf(buf, "LDI R%d,%ld", ri, d20);
	return buf;
    case MVM_LDHI:
	sprintf(buf, "LDHI R%d,%ld", ri, d20);
	return buf;
    case MVM_POP:
	sprintf(buf, "POP R%d", ri);
	return buf;
    case MVM_PUSH:
	sprintf(buf, "PUSH R%d", ri);
	return buf;
    case MVM_CALL:
	sprintf(buf, "CALL R%d,%ld", ri, d20);
	return buf;
    case MVM_JPL:
	sprintf(buf, "JPL R%d,%ld", ri, d20);
	return buf;
    case MVM_IN:
	sprintf(buf, "IN R%d", ri);
	return buf;
    case MVM_OUT:
	sprintf(buf, "OUT R%d", ri);
	return buf;
    case MVM_INQ:
	sprintf(buf, "INQ R%d", ri);
	return buf;
    case MVM_OUTQ:
	sprintf(buf, "OUTQ R%d", ri);
	return buf;
    case MVM_OUTS:
	sprintf(buf, "OUTS R%d,%ld", ri, d20);
	return buf;
    case MVM_EXIT:
	sprintf(buf, "EXIT R%d", ri);
	return buf;
    case MVM_JP:
	sprintf(buf, "JP R%d,%ld", ri, d20);
	return buf;
    case MVM_JPE:
	sprintf(buf, "JPE R%d,%ld", ri, d20);
	return buf;
    case MVM_JPNE:
	sprintf(buf, "JPNE R%d,%ld", ri, d20);
	return buf;
    case MVM_JPGT:
	sprintf(buf, "JPGT R%d,%ld", ri, d20);
	return buf;
    case MVM_JPLE:
	sprintf(buf, "JPLE R%d,%ld", ri, d20);
	return buf;
    case MVM_JPLT:
	sprintf(buf, "JPLT R%d,%ld", ri, d20);
	return buf;
    case MVM_JPGE:
	sprintf(buf, "JPGE R%d,%ld", ri, d20);
	return buf;
    case MVM_JPGTU:
	sprintf(buf, "JPGTU R%d,%ld", ri, d20);
	return buf;
    case MVM_JPLEU:
	sprintf(buf, "JPLEU R%d,%ld", ri, d20);
	return buf;
    case MVM_JPC:
	sprintf(buf, "JPC R%d,%ld", ri, d20);
	return buf;
    case MVM_JPNC:
	sprintf(buf, "JPNC R%d,%ld", ri, d20);
	return buf;
    case MVM_JPNEG:
	sprintf(buf, "JPNEG R%d,%ld", ri, d20);
	return buf;
    case MVM_JPNNEG:
	sprintf(buf, "JPNNEG R%d,%ld", ri, d20);
	return buf;
    case MVM_JPV:
	sprintf(buf, "JPV R%d,%ld", ri, d20);
	return buf;
    case MVM_JPNV:
	sprintf(buf, "JPNV R%d,%ld", ri, d20);
	return buf;
    case MVM_ADD:
	sprintf(buf, "ADD R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_ADDC:
	sprintf(buf, "ADDC R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_SUB:
	sprintf(buf, "SUB R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_SUBC:
	sprintf(buf, "SUBC R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_AND:
	sprintf(buf, "AND R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_OR:
	sprintf(buf, "OR R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_XOR:
	sprintf(buf, "XOR R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_NXOR:
	sprintf(buf, "NXOR R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_SLL:
	sprintf(buf, "SLL R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_SRL:
	sprintf(buf, "SRL R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_SRA:
	sprintf(buf, "SRA R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_ADDI:
	sprintf(buf, "ADDI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_ADDIC:
	sprintf(buf, "ADDIC R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_SUBI:
	sprintf(buf, "SUBI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_SUBIC:
	sprintf(buf, "SUBIC R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_ANDI:
	sprintf(buf, "ANDI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_ORI:
	sprintf(buf, "ORI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_XORI:
	sprintf(buf, "XORI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_NXORI:
	sprintf(buf, "NXORI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_SLLI:
	sprintf(buf, "SLLI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_SRLI:
	sprintf(buf, "SRLI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_SRAI:
	sprintf(buf, "SRAI R%d,R%d,%ld", ri, rj, k16);
	return buf;
    case MVM_MUL:
	sprintf(buf, "MUL R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_DIV:
	sprintf(buf, "DIV R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_MOD:
	sprintf(buf, "MOD R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_MULU:
	sprintf(buf, "MULU R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_DIVU:
	sprintf(buf, "DIVU R%d,R%d,R%d", ri, rj, rk);
	return buf;
    case MVM_MODU:
	sprintf(buf, "MODU R%d,R%d,R%d", ri, rj, rk);
	return buf;
    default:
	sprintf(buf, "̿ 0x%08lx", inst);
	return buf;
    }
}

static void Dump(void)
{
    int i, j, d, p, m, n;

    printf ("      ץ        |  쥸   |  å\n");
    m = MVM_NREG + 1;
    n = ((MVM_MEMLEN - SP)/sizeof(Word)+m-1)/m;
    p = -4;
    if (p < - (int)(PC/sizeof(Word)))
	p = - (int)(PC/sizeof(Word));
    for (i = 0; i < m; i ++) {
	printf("%c%4ld %-18s", (p + i == 0) ? '>' : ' ',
	       PC + (p + i) * sizeof(Word), Inst(Mem[PC/sizeof(Word) + p + i]));

	switch (i) {
	case 0: case 1: case 2: case 3: case 4: case 5:
	case 6: case 7: case 8: case 9: case 10: case 12:
	    printf ("| %sR%d=%-11ld |", (i<10) ? " " : "", i, Reg[i]);
	    break;
	case 11:
	    printf ("|  GP=%-11ld |",  Reg[i]);
	    break;
	case 13:
	    printf ("|  FP=%-11ld |",  Reg[i]);
	    break;
	case 14:
	    printf ("|  SP=%-11ld |",  Reg[i]);
	    break;
	case 15:
	    printf ("|  PC=%-11ld |",  Reg[i]);
	    break;
	case 16:
	    printf ("|  ZNVC=%d%d%d%d      |", Z, N, V, C);
	    break;
	default:
	    printf ("| %15s |", "");
	    break;
	}

	for (j = 0; j < n && j < 2; j ++) {
	    d = MVM_MEMLEN + ((j-n)*m+i)*sizeof(Word) - SP;
	    if (0 <= d) {
		d += SP-FP;
		if (d == 0)
		    printf ("     FP:%-9ld", Mem[(FP+d)/sizeof(Word)]);
		else if (d < 0)
		    printf (" %sFP%+d:%-9ld",
			    (d<-99) ? "" : (d<-9) ? " " : "  ",
			    d, Mem[(FP+d)/sizeof(Word)]);
		else
		    printf (" %sFP%+d:%-9ld", (d>99) ? "" : (d>9) ? " " : "  ",
			    d, Mem[(FP+d)/sizeof(Word)]);
	    }
	    else if (j < n - 1)
		printf ("        %9s", "");
	}
	printf("\n");
    }
}

static void ReadObjectFile(char * filename)
{
    FILE *fp;
    Address addr = 0;
    int c, i;

    fp = fopen (filename, "rb");
    if (fp == (FILE *)0) {
	fprintf (stderr, "ե뤬ץǤޤ : %s\n", filename);
	exit(1);
    }

    memset((void *)Mem, 0, sizeof(Mem));
    while (addr < MVM_MEMLEN) {
	Mem[addr/sizeof(Word)] = 0;
	for (i = sizeof(Word) - 1; i >= 0; i--) {
	    if ((c = fgetc(fp)) == EOF)
		break;
	    Mem[addr/sizeof(Word)] |= c << (i*8);
	}
	if (i >= 0)
	    break;
	if (Disassemble)
	    printf("%8ld  %s\n", addr, Inst(Mem[addr/sizeof(Word)]));
	addr += sizeof(Word);
    }

    fclose(fp);
    if (Disassemble)
	exit(0);
}

#define MSB(x)      (((x) & (1 << 31)) != 0)

static Word GetMem(Address addr)
{
    if (addr < 0 || (addr % sizeof(Word)) != 0
	    || MVM_MEMLEN < addr+sizeof(Word)) {
	fprintf (stderr, "ǡɤ߽Фɥ쥹 (%ld) Ǥ\n", addr);
	PC -= sizeof(Word);
	if (!Step)
	    Dump();
	exit(1);
    }
    return Mem[addr/sizeof(Word)];
}

static void SetMem(Address addr, Word val)
{
    if (addr < 0 || (addr % sizeof(Word)) != 0
	    || MVM_MEMLEN < addr+sizeof(Word)) {
	fprintf (stderr, "ǡ񤭹ॢɥ쥹 (%ld) Ǥ\n", addr);
	PC -= sizeof(Word);
	if (!Step)
	    Dump();
	exit(1);
    }
    Mem[addr/sizeof(Word)] = val;
}

static Word Add(Word operand1, Word operand2, Word carry)
{
    Word result = operand1 + operand2 + carry;

    Z = (result == 0);
    N = MSB(result);
    C = ((MSB(operand1) && MSB(operand2))
	    || ((MSB(operand1) || MSB(operand2)) && !MSB(result)));
    V = ((MSB(operand1) && MSB(operand2) && !MSB(result))
	    || (!MSB(operand1) && !MSB(operand2) && MSB(result)));
    return result;
}

static Word Sub(Word operand1, Word operand2, Word carry)
{
    Word result = operand1 - operand2 - carry;

    Z = (result == 0);
    N = MSB(result);
    C = ((!MSB(operand1) && MSB(operand2))
	    || ((!MSB(operand1) || MSB(operand2)) && MSB(result)));
    V = ((MSB(operand1) && !MSB(operand2) && !MSB(result))
	    || (!MSB(operand1) && MSB(operand2) && MSB(result)));
    return result;
}

static Word Logical(Word result)
{
    Z = (result == 0);
    N = MSB(result);
    C = V = 0;
    return result;
}

static Word SRA(Word operand1, Word operand2)
{
    Word result = (operand1 >> operand2);

    if (MSB(operand1)) {
	if (operand2 < 32)
	    result = (result | (~0 << (32 - operand2)));
	else
	    result = ~0;
    }
    return result;
}

static void Execute (void)
{
    Word inst;
    Address d20, k16, addr;
    int ri, rk, rj;

    Z = N = C = V = 0;
    memset((void *)Reg, 0, sizeof(Reg));
    SP = MVM_MEMLEN;

    while (1) {
	if ((PC & (sizeof(Word)-1)) != 0 || MVM_MEMLEN <= PC) {
	    fprintf (stderr, "¹Ԥ륢ɥ쥹 (%ld) Ǥ\n", PC);
	    if (!Step)
		Dump();
	    exit(1);
	}

	inst = Mem[PC/sizeof(Word)];
	Reg[MVM_R0] = 0;

	if (Step) {
	    Dump();
	    printf ("\n꥿󥭡򲡤Ʋ\n");
	    fflush (stdout);
	    fflush (stdin);
	    getchar ();
	}

	PC += sizeof(Word);

	d20 = (inst & 0xfffff);
	if (inst & 0x80000)
	    d20 |= ~0xfffff;
	ri = (inst >> 24) & 0xf;
	addr = Reg[(inst >> 20) & 0xf] + d20;

	switch ((inst >> 24) & 0xf0) {
	case MVM_LD:
	    Reg[ri] = GetMem(addr);
	    continue;
	case MVM_ST:
	    SetMem(addr, Reg[ri]);
	    continue;
	case MVM_LDB:
	    Reg[ri] = (Mem[addr/sizeof(Word)]
			>> (8*(sizeof(Word) - 1 - addr % sizeof(Word)))) & 0xff;
	    if (Reg[ri] & 0x80)
		Reg[ri] |= ~0xff;
	    continue;
	case MVM_LDUB:
	    Reg[ri] = (Mem[addr/sizeof(Word)]
			>> (8*(sizeof(Word) - 1 - addr % sizeof(Word)))) & 0xff;
	    continue;
	case MVM_STB:
	    Mem[addr/sizeof(Word)]
		&= ~(0xff << (8*(sizeof(Word) - 1 - addr % sizeof(Word))));
	    Mem[addr/sizeof(Word)]
		|= ((Reg[ri] & 0xff)
			<< (8*(sizeof(Word) - 1 - addr % sizeof(Word))));
	    continue;
	}

	k16 = (inst & 0xffff);
	ri = (inst >> 20) & 0xf;
	rj = (inst >> 16) & 0xf;
	rk = k16 & 0xf;

	switch ((inst >> 24) & 0xff) {
	case MVM_LDI:
	    Reg[ri] = d20;
	    continue;
	case MVM_LDHI:
	    Reg[ri] = (d20 << 16);
	    continue;
	case MVM_POP:
	    Reg[ri] = GetMem(SP);
	    SP += sizeof(Word);
	    continue;
	case MVM_PUSH:
	    SP -= sizeof(Word);
	    SetMem(SP, Reg[ri]);
	    continue;
	case MVM_CALL:
	    SP -= sizeof(Word);
	    SetMem(SP, PC);
	    PC = addr;
	    continue;
	case MVM_JPL:
	    Reg[MVM_RA] = PC;
	    PC = addr;
	    continue;
	case MVM_IN:
	    printf("? ");
	    scanf("%ld", &Reg[ri]);
	    continue;
	case MVM_OUT:
	    printf("%ld\n", Reg[ri]);
	    continue;
	case MVM_INQ:
	    scanf("%ld", &Reg[ri]);
	    continue;
	case MVM_OUTQ:
	    printf("%ld", Reg[ri]);
	    continue;
	case MVM_OUTS:
	    printf("%s", (char *)Mem + (unsigned int)addr);
	    continue;
	case MVM_EXIT:
	    exit(Reg[ri]);
	case MVM_JP:
	    PC = addr;
	    continue;
	case MVM_JPE:
	    if (Z == 1) PC = addr;
	    continue;
	case MVM_JPNE:
	    if (Z == 0) PC = addr;
	    continue;
	case MVM_JPGT:
	    if (N == V && Z == 0) PC = addr;
	    continue;
	case MVM_JPLE:
	    if (N != V || Z == 1) PC = addr;
	    continue;
	case MVM_JPLT:
	    if (N != V) PC = addr;
	    continue;
	case MVM_JPGE:
	    if (N == V) PC = addr;
	    continue;
	case MVM_JPGTU:
	    if (C == 0 && Z == 0) PC = addr;
	    continue;
	case MVM_JPLEU:
	    if (C == 1 || Z == 1) PC = addr;
	    continue;
	case MVM_JPC:
	    if (C == 1) PC = addr;
	    continue;
	case MVM_JPNC:
	    if (C == 0) PC = addr;
	    continue;
	case MVM_JPNEG:
	    if (N == 1) PC = addr;
	    continue;
	case MVM_JPNNEG:
	    if (N == 0) PC = addr;
	    continue;
	case MVM_JPV:
	    if (V == 1) PC = addr;
	    continue;
	case MVM_JPNV:
	    if (V == 0) PC = addr;
	    continue;
	case MVM_ADD:
	    Reg[ri] = Add(Reg[rj], Reg[rk], 0);
	    continue;
	case MVM_ADDC:
	    Reg[ri] = Add(Reg[rj], Reg[rk], 1);
	    continue;
	case MVM_SUB:
	    Reg[ri] = Sub(Reg[rj], Reg[rk], 0);
	    continue;
	case MVM_SUBC:
	    Reg[ri] = Sub(Reg[rj], Reg[rk], 1);
	    continue;
	case MVM_AND:
	    Reg[ri] = Logical(Reg[rj] & Reg[rk]);
	    continue;
	case MVM_OR:
	    Reg[ri] = Logical(Reg[rj] | Reg[rk]);
	    continue;
	case MVM_XOR:
	    Reg[ri] = Logical(Reg[rj] ^ Reg[rk]);
	    continue;
	case MVM_NXOR:
	    Reg[ri] = Logical(~(Reg[rj] ^ Reg[rk]));
	    continue;
	case MVM_SLL:
	    Reg[ri] = Reg[rj] << Reg[rk];
	    continue;
	case MVM_SRL:
	    Reg[ri] = Reg[rj] >> Reg[rk];
	    continue;
	case MVM_SRA:
	    Reg[ri] = SRA(Reg[rj], Reg[rk]);
	    continue;
	case MVM_ADDI:
	    Reg[ri] = Add(Reg[rj], k16, 0);
	    continue;
	case MVM_ADDIC:
	    Reg[ri] = Add(Reg[rj], k16, 1);
	    continue;
	case MVM_SUBI:
	    Reg[ri] = Sub(Reg[rj], k16, 0);
	    continue;
	case MVM_SUBIC:
	    Reg[ri] = Sub(Reg[rj], k16, 1);
	    continue;
	case MVM_ANDI:
	    Reg[ri] = Logical(Reg[rj] & k16);
	    continue;
	case MVM_ORI:
	    Reg[ri] = Logical(Reg[rj] | k16);
	    continue;
	case MVM_XORI:
	    Reg[ri] = Logical(Reg[rj] ^ k16);
	    continue;
	case MVM_NXORI:
	    Reg[ri] = Logical(~(Reg[rj] ^ k16));
	    continue;
	case MVM_SLLI:
	    Reg[ri] = Reg[rj] << k16;
	    continue;
	case MVM_SRLI:
	    Reg[ri] = Reg[rj] >> k16;
	    continue;
	case MVM_SRAI:
	    Reg[ri] = SRA(Reg[rj], k16);
	    continue;
	case MVM_MUL:
	    Reg[ri] = (SWord)Reg[rj] * (SWord)Reg[rk];
	    continue;
	case MVM_DIV:
	    Reg[ri] = (SWord)Reg[rj] / (SWord)Reg[rk];
	    continue;
	case MVM_MOD:
	    Reg[ri] = (SWord)Reg[rj] % (SWord)Reg[rk];
	    continue;
	case MVM_MULU:
	    Reg[ri] = Reg[rj] * Reg[rk];
	    continue;
	case MVM_DIVU:
	    Reg[ri] = Reg[rj] / Reg[rk];
	    continue;
	case MVM_MODU:
	    Reg[ri] = Reg[rj] % Reg[rk];
	    continue;
	default:
	    fprintf (stderr, "̿¹Ԥ褦Ȥޤ\n");
	    PC -= sizeof(Word);
	    if (!Step)
		Dump();
	    exit(1);
	}
    }
}

int main (argc, argv)
    int argc;
    char *argv[];
{
    char *filename = "mcc.out";

    if (argc > 1) {
	if (strcmp (argv[1], "-d") == 0) {
	    Disassemble = 1;
	    argc --;
	    argv ++;
	}
	else if (strcmp (argv[1], "-t") == 0) {
	    Step = 1;
	    argc --;
	    argv ++;
	}
    }

    if (argc > 1)
	filename = argv[1];

    ReadObjectFile(filename);
    Execute();
    exit(0);
}

