/*
 *      parse.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "scan.h"
#include "gen.h"

typedef enum { Var, Fun, Unknown } ObjectType;

typedef struct Object {
    ObjectType      type;       /* ֥Ȥμ(ѿ/ؿ/̤) */
    char            name[MC_MAXIDLEN+1];        /* ѿ̾/ؿ̾ */
    int             reg;        /* ѿɥ쥹δȤʤ쥸ֹ
				   ؿξ R0ѿξ GP (R11)
				   ؿβɽѿξ FP (R13) */
    Offset          offset;     /* reg Υեå(а) */
    struct Object * next;       /* ƱפμΥ֥ */
} Object;

typedef struct Scope {
    struct Scope *  outer;      /* ĳ¦Υ */
    Object *        objects;    /* ΥפΥ֥ȤΥꥹ */
    Offset          offset;     /* ѿμγ() */
} Scope;

static Scope * GlobalScope;     /* (ֳ¦) */
static Scope * CurrentScope;    /* ߥѥΥ */

/* FindObject() : ΥפΥ֥ꥹȤ饪֥Ȥõ
		  Ф Ĥʤ 0 ֤ */

static Object * FindObject(Scope *scope, char *name)
{
    Object *obj;

    for (obj = scope->objects; obj != (Object *)0; obj = obj->next) {
	if (strcmp(name, obj->name) == 0)
	    return obj;
    }
    return (Object *)0;
}

/* NewObject() : ֥Ȥ롣 */

static Object * NewObject(ObjectType type, char *name)
{
    Object *obj;

    obj = (Object *) malloc(sizeof(Object));
    if (obj == (Object *)0)
	Error("ѥ뤹뤿Υ꤬­ޤ");
    strcpy(obj->name, name);
    obj->type = type;
    obj->next = (Object *)0;
    return obj;
}

/* NewVar() : ѿ򸽺ߤΥפɲä롣*/

static Object * NewVar(char *name)
{
    Object *obj;

    if (FindObject(CurrentScope, name) != (Object *)0)
	Error("ѿ̾ޤϰ̾ͤƤޤ");

    obj = NewObject(Var, name);
    obj->next = CurrentScope->objects;

    if (CurrentScope->outer == (Scope *)0) { /* ѿ */
	obj->reg = MVM_GP;
	obj->offset = CurrentScope->offset;
	CurrentScope->offset += sizeof(Word);
    }
    else if (CurrentScope->outer->outer == (Scope *)0) { /*  */
	obj->reg = MVM_FP;
	obj->offset = 0;        /*  FunDef() ꤹ */
    }
    else { /* ɽѿ */
	obj->reg = MVM_FP;
	CurrentScope->offset -= sizeof(Word);
	obj->offset = CurrentScope->offset;
    }

    CurrentScope->objects = obj;
    return obj;
}

/* UseObject() : ֥ȤõФ ĤʤС
		 type=Unknown Ȥ襹פ˥֥Ȥ롣 */

static Object * UseObject(char *name)
{
    Scope *scope;
    Object *obj;

    /* ¦Υפ襹פ˸äƽ˥֥Ȥõ */
    for (scope = CurrentScope; scope != (Scope *)0; scope = scope->outer) {
	obj = FindObject(scope, name);
	if (obj != (Object *)0)
		return obj;
    }

    /* 襹פ˿֥Ȥ */
    obj = NewObject(Unknown, name);
    obj->reg = MVM_R0;
    obj->offset = 0;
    obj->next = GlobalScope->objects;
    GlobalScope->objects = obj;
    return obj;
}

/* NewScope() : פ CurrentScope Ȥ롣 */

static void NewScope(Offset offset)
{
    Scope * scope;

    scope = (Scope *) malloc(sizeof(Scope));
    if (scope == (Scope *)0)
	Error("ѥ뤹뤿Υ꤬­ޤ");
    scope->outer = CurrentScope;
    scope->objects = (Object *)0;
    scope->offset = offset;
    CurrentScope = scope;
}

/* ExitScope() : CurrentScope Ѵĳ¦Υפ˽Ф롣 */

static void ExitScope(void)
{
    char message[100+MC_MAXIDLEN];
    Scope *cur = CurrentScope;
    Object *obj;

    /* ̤ؿΥå */
    for (obj = cur->objects; obj != (Object *)0; obj = obj->next) {
	if (obj->type == Unknown) {
	    sprintf(message, "ؿ %s ̤Ǥ", obj->name);
	    Error(message);
	}
    }

    CurrentScope = CurrentScope->outer;
    free((void *)cur);
}

#define N_EXPREG 4      /* ͤη׻˻Ѥʪ쥸 */

/* RealReg(): ۥ쥸 vreg Ƥ٤ʪ쥸ֹ */
#define RealReg(vreg)   (MVM_R1 + (vreg) % N_EXPREG)

static int CurrentVRegTop = 0;  /* βۥ쥸κֹ */

/* ActivateVReg(): ۥ쥸 vreg Ѳǽʾ֤ˤ롣 ʪ쥸
		   Ǥ˰դʤС򥹥å򤷤ʪ
		    vreg ȤƻѤǤ褦ˤ롣 ޤvreg
		   Ǥ˥å򤵤Ƥϥå夫
		   ʪ쥸롣 */

static void ActivateVReg(int vreg)
{
    while (CurrentVRegTop < vreg) {
	CurrentVRegTop ++;
	if (N_EXPREG <= CurrentVRegTop)
	    PutInst(MVM_PUSH, RealReg(CurrentVRegTop), 0, 0);
    }

    while (vreg <= CurrentVRegTop - N_EXPREG) {
	PutInst(MVM_POP, RealReg(CurrentVRegTop), 0, 0);
	CurrentVRegTop --;
    }
}

/* SaveVReg(): ؿƤӽФäơvreg 꾮ֹβۥ쥸
	       ˻ȤƤʪ쥸򤹤٤ƥå򤹤롣 */

static void SaveVReg(int vreg)
{
    int r, oldest = CurrentVRegTop - N_EXPREG + 1;

    if (oldest < 0)
	oldest = 0;
    for (r = oldest; r < vreg; r ++)
	PutInst(MVM_PUSH, RealReg(r), 0, 0);
    CurrentVRegTop = 0;
}

/* RestoreVReg(): SaveVReg(vreg) ˤäƲۥ쥸å򤵤
		  ξ礹褦 CurrentVRegTop ꤹ롣
		  RestoreVReg() ϴؿƤӽФθ˸ƤӽФ롣 */

static void RestoreVReg(int vreg)
{
    CurrentVRegTop = vreg + N_EXPREG - 1;
}

static void Exp(int vreg);
static void Block(void);


/* FunArgs(): ؿƤӽФμ°ʸϤΰ֤˼
	      ͤ׻ƥå PUSH ̿ᷲ롣
	      FunArgs() ֤ͤϼ°θĿ */

static int FunArgs(void)
{
    int n = 0;

    if (token != mc_lparen)
	Error("ؿƽФΰꥹȤ ( ޤ");
    GetToken();
    if (token != mc_rparen) {
	Exp(0);                 /* ǽμ° */
	PutInst(MVM_PUSH, RealReg(0), 0, 0);
	n ++;
	while (token == mc_comma) {
	    GetToken();
	    Exp(0);             /* 2ܰʹߤμ° */
	    PutInst(MVM_PUSH, RealReg(0), 0, 0);
	    n ++;
	}
	if (token != mc_rparen)
	    Error("ؿƽФΰꥹȤ ) ޤ");
    }
    GetToken();
    return n;
}

/* Factor(): <Factor> ʸϤͤ׻̤ۥ쥸
	      vreg ֤褦̿롣 */

static void Factor(int vreg)
{
    int n;
    Address l1;
    Object *obj;

    switch (token) {
    case mc_plusplus: case mc_minusminus:
        /* ʬ黻: <IncDef><Ident> */
        GetToken();
    case mc_ident:
	/* ѿ: <Ident>ޤϡؿƤӽФ: <Ident> <FunArgs> */
	obj = UseObject(ident);
	GetToken();
	if (token == mc_lparen) {
	    /* ؿƤӽФ: <Ident> <FunArgs> */
	    if (obj->type == Var)
		Error("ѿФؿƽФϤǤޤ");
	    SaveVReg(vreg);
	    n = FunArgs();
	    l1 = CurrentPC;
	    PutInst(MVM_CALL, obj->reg, 0, obj->offset);
	    if (obj->type == Unknown) {
		/* ̤δؿʤΤǡ CALL ̿ꥹȤɲä롣
		   δؿ줿 CALL ̿Υɥ쥹ʬ
		   ꤹ롣 */
		obj->offset = l1;
	    }
	    if (n > 0)
		PutInst(MVM_ADDI, MVM_SP, MVM_SP, n*sizeof(Word));
	    if (RealReg(vreg) != RealReg(0))
		PutInst(MVM_ADD, RealReg(vreg), RealReg(0), MVM_R0);
	    RestoreVReg(vreg);
	}
	else if (obj->type == Var) {
	    /* ѿ: <Ident> */
	    ActivateVReg(vreg);
	    PutInst(MVM_LD, RealReg(vreg), obj->reg, obj->offset);
	}
	else
	    Error("ƤʤѿϻѤǤޤ");
	break;
    case mc_number:
	/* : <Number> */
	ActivateVReg(vreg);
	if (number < (1 << 19)) {
	    PutInst(MVM_LDI, RealReg(vreg), 0, number);
	}
	else {
	    PutInst(MVM_LDHI, RealReg(vreg), 0, number >> 16);
	    PutInst(MVM_ADDI, RealReg(vreg), RealReg(vreg),
		    number & 0xffff);
	}
	GetToken();
	break;
    case mc_lparen:
	/* ̤ǰϤޤ줿: ( <Exp> ) */
	GetToken();
	Exp(vreg);
	if (token == mc_rparen)
	    GetToken();
	else
	    Error("Ĥ̤٤֤ˤޤ");
	break;
    default:
	Error("ʸˡ꤬ޤ");
	break;
    }
}

/* Term(): <Term> ʸϤͤ׻̤ۥ쥸 vreg
	   ֤褦̿롣 */

static void Term(int vreg)
{
    Token op;

    Factor(vreg);
    while (token == mc_mult || token == mc_div || token == mc_mod) {
	op = token;
	GetToken();
	Factor(vreg+1);
	ActivateVReg(vreg);
	switch (op) {
	case mc_mult:
	    PutInst(MVM_MUL, RealReg(vreg), RealReg(vreg), RealReg(vreg+1));
	    break;
	case mc_div:
	    PutInst(MVM_DIV, RealReg(vreg), RealReg(vreg), RealReg(vreg+1));
	    break;
	case mc_mod:
	    PutInst(MVM_MOD, RealReg(vreg), RealReg(vreg), RealReg(vreg+1));
	    break;
	}
    }
}

/* Exp(): <Exp> ʸϤͤ׻̤ۥ쥸 vreg 
	  褦̿롣 */

static void Exp(int vreg)
{
    Token op = mc_plus;

    if (token == mc_plus || token == mc_minus) {
	op = token;
	GetToken();
    }
    Term(vreg);
    if (op == mc_minus)
	PutInst(MVM_SUB, RealReg(vreg), MVM_R0, RealReg(vreg));
    while (token == mc_plus || token == mc_minus) {
	op = token;
	GetToken();
	Term(vreg+1);
	ActivateVReg(vreg);
	if (op == mc_minus)
	    PutInst(MVM_SUB, RealReg(vreg), RealReg(vreg), RealReg(vreg+1));
	else
	    PutInst(MVM_ADD, RealReg(vreg), RealReg(vreg), RealReg(vreg+1));
    }
}

/* Cond(): <Cond> ʸϤξ郎ΩʤʬԤ褦
	   ̿롣 ʬ̿᤬֤줿ɥ쥹֤ */

static Address Cond(void)
{
    Address l;
    Word op;
    Token rel;

    Exp(0);
    if (mc_eq <= token && token <= mc_ge) {
	rel = token;
	GetToken();
	Exp(1);
	PutInst(MVM_SUB, MVM_R0, RealReg(0), RealReg(1));
	l = CurrentPC;
	switch (rel) {
	    case mc_eq: op = MVM_JPNE; break;
	    case mc_neq: op = MVM_JPE; break;
	    case mc_gt: op = MVM_JPLE; break;
	    case mc_ge: op = MVM_JPLT; break;
	    case mc_lt: op = MVM_JPGE; break;
	    case mc_le: op = MVM_JPGT; break;
	}
	PutInst(op, MVM_R0, 0, 0);
	return l;
    }
    else
	Error("Ｐʸˡ꤬ޤ");
}

/* Statement(): <Statement> ʸϤб̿롣*/

static void Statement(void)
{
    int n;
    Address l1, l2;
    Object *obj;

    switch (token) {
    case mc_ident:
	/* <Ident> = <Exp>; | <Ident> <FunArgs>; */
	obj = UseObject(ident);
	GetToken();
	switch (token) {
	case mc_assign:
	    if (obj->type != Var)
		Error("ƤʤѿؤϤǤޤ");
	    GetToken();
	    Exp(0);
	    PutInst(MVM_ST, RealReg(0), obj->reg, obj->offset);
	    if (token != mc_semicolon)
		Error("ʸ ; ޤ");
	    GetToken();
	    break;
	case mc_lparen:
	    if (obj->type == Var)
		Error("ѿФؿƽФϤǤޤ");
	    n = FunArgs();
	    l1 = CurrentPC;
	    PutInst(MVM_CALL, obj->reg, 0, obj->offset);
	    if (obj->type == Unknown)
		obj->offset = l1;
	    PutInst(MVM_ADDI, MVM_SP, MVM_SP, n*sizeof(Word));
	    if (token != mc_semicolon)
		Error("ؿƽФʸ ; ޤ");
	    GetToken();
	    break;
	default:
	    Error("ʸ =ޤϡؿƽФʸ ( ޤ");
	    break;
	}
	break;
    case mc_return:
	/* return <Exp>; */
	GetToken();
	Exp(0);
	if (CurrentScope->offset < 0)
	    PutInst(MVM_ADDI, MVM_SP, MVM_FP, MVM_R0);
	PutInst(MVM_POP, MVM_FP, 0, 0);
	PutInst(MVM_POP, MVM_PC, 0, 0);
	if (token != mc_semicolon)
	    Error("return ʸ ; ޤ");
	GetToken();
	break;
    case mc_input:
	/* input <Ident>; */
	GetToken();
	if (token != mc_ident)
	    Error("input ʸѿ̾ޤ");
	obj = UseObject(ident);
	PutInst(MVM_IN, MVM_R1, 0, 0);
	PutInst(MVM_ST, MVM_R1, obj->reg, obj->offset);
	GetToken();
	if (token != mc_semicolon)
	    Error("input ʸ ; ޤ");
	GetToken();
	break;
    case mc_print:
	/* print <Exp>; */
	GetToken();
	Exp(0);
	PutInst(MVM_OUT, RealReg(0), 0, 0);
	if (token != mc_semicolon)
	    Error("print ʸ ; ޤ");
	GetToken();
	break;
    case mc_if:
	/* if ( <Cond> ) <Statement> [ else <Statement> ] */
	GetToken();
	if (token != mc_lparen)
	    Error("if θ ( ޤ");
	GetToken();
	l1 = Cond();
	if (token != mc_rparen)
	    Error("if ξＰθ ) ޤ");
	GetToken();
	Statement();
	if (token == mc_else) {
	    l2 = CurrentPC;
	    PutInst(MVM_JP, MVM_R0, 0, 0);
	    FixInst(l1, CurrentPC);
	    GetToken();
	    Statement();
	    FixInst(l2, CurrentPC);
	}
	else
	    FixInst(l1, CurrentPC);
	break;
    case mc_while:
	/* while ( <Cond> ) <Statement> */
	GetToken();
	if (token != mc_lparen)
	    Error("while θ ( ޤ");
	GetToken();
	l1 = CurrentPC;
	l2 = Cond();
	if (token != mc_rparen)
	    Error("while ξＰθ ) ޤ");
	GetToken();
	Statement();
	PutInst(MVM_JP, MVM_R0, 0, l1);
	FixInst(l2, CurrentPC);
	break;
    case mc_lcb:
	/* "{" [ <VarDef> ] { <Statement> } "}" */
	Block();
	break;
    case mc_semicolon:
	Error("ʸϵƤޤ");
	break;
    default:
	Error("ʸϤ٤˸꤬ޤ");
	break;
    }
}

/* VarDef(): <VarDef> ʸϤCurrentScope Υ֥ȥꥹȤ
	     롣 */

static int VarDef(void)
{
    int n = 0;

    GetToken();
    if (token != mc_ident)
	Error("ѿ̾ޤ");
    NewVar(ident);
    n ++;
    GetToken();
    while (token == mc_comma) {
	GetToken();
	if (token == mc_ident) {
	    NewVar(ident);
	    n ++;
	    GetToken();
	}
    }
    if (token != mc_semicolon)
	Error("ѿ ,  ; ޤ");
    GetToken();
    return n;
}

/* FunDef(): <FunDef> ʸϤؿб̿롣 */

static void FunDef(void)
{
    Offset offset;
    Object *obj;

    obj = UseObject(ident);
    if (obj->type == Var)
	Error("ѿ̾ؿ̾Ȥƺ뤳ȤϤǤޤ");
    if (obj->type == Fun)
	Error("ؿ뤳ȤϤǤޤ");
    obj->type = Fun;
    GetToken();

    /* δؿФƤǤ줿 CALL ̿Υɥ쥹ʬꤹ롣 */
    FixInstChain(obj->offset, CurrentPC);

    /* δؿ֥ȤΥɥ쥹ꤹ롣 */
    obj->offset = CurrentPC;

    /* פɲä򤽤Υ֥ȥꥹȤϿ롣 */
    NewScope((Offset)0);
    if (token != mc_lparen)
	Error("ؿΰꥹȤ ( ޤ");
    GetToken();
    if (token != mc_rparen) {
	if (token != mc_ident)
	    Error("ؿΰꥹȤʸˡ꤬ޤ");
	obj = NewVar(ident);
	GetToken();
	while (token == mc_comma) {
	    GetToken();
	    if (token != mc_ident)
		Error("ؿΰꥹȤʸˡ꤬ޤ");
	    obj = NewVar(ident);
	    GetToken();
	}
	if (token != mc_rparen)
	    Error("ؿΰꥹȤ ) ޤ");
    }
    GetToken();

    /* Υɥ쥹(offset)׻ꤹ롣 */
    offset = 2 * sizeof(Word);  /* FP ǸΰޤǤΥեå */
    /* ǸΰˡΥեåȤ */
    for (obj = CurrentScope->objects; obj != (Object *)0; obj = obj->next) {
	obj->offset = offset;
	offset += sizeof(Word);
    }

    /* ̿ */
    PutInst(MVM_PUSH, MVM_FP, 0, 0);
    PutInst(MVM_ADD, MVM_FP, MVM_SP, MVM_R0);
    Block();
    PutInst(MVM_POP, MVM_FP, 0, 0);
    PutInst(MVM_POP, MVM_PC, 0, 0);
    ExitScope();
}

/* Block(): <Block> ʸϤб̿롣 */

static void Block(void)
{
    int n = 0;

    /* פɲäɽѿ򤽤Υ֥ȥꥹȤϿ롣 */
    NewScope(CurrentScope->offset);
    GetToken();
    if (token == mc_int) {
	n = VarDef();
	/* ɽѿο˸礦ΰ򥹥å˳ݤ롣 */
	PutInst(MVM_SUBI, MVM_SP, MVM_SP, n*sizeof(Word));
    }

    /* ֥åγ <Statement> 롣*/
    while (token != mc_rcb)
	Statement();

    /* åݥ󥿤ᤷơɽѿΤΰ롣 */
    if (n > 0)
	PutInst(MVM_ADDI, MVM_SP, MVM_SP, n*sizeof(Word));

    /* Υ֥åΤΥפѴ롣 */
    GetToken();
    ExitScope();
}

/* Prog(): ץ <Prog> ʸϤб̿롣 */

static void Prog(void)
{
    Address l1, l2;
    Object *obj;

    /* νԤ */
    InitInst();
    CurrentVRegTop = 0;

    /* 襹פ */
    CurrentScope = (Scope *)0;
    NewScope((Offset)0);
    GlobalScope = CurrentScope;

    /* ɤ */
    PutInst(MVM_ADD, MVM_FP, MVM_SP, MVM_R0); /* ե졼ݥ󥿤ν */
    l1 = CurrentPC;
    PutInst(MVM_LDI, MVM_GP, 0, 0); /* Хݥ󥿤ν */
    obj = UseObject("main");
    l2 = CurrentPC;
    PutInst(MVM_CALL, obj->reg, 0, obj->offset); /* main θƤӽФ */
    obj->offset = l2;
    PutInst(MVM_EXIT, MVM_R0, 0, 0); /* ץνλ̿ */

    /* ѿ <VarDef>ؿ <FunDef> ν */
    while (token == mc_int || token == mc_ident) {
	if (token == mc_int)
	    VarDef();
	else
	    FunDef();
    }

    /* Хݥ󥿤νͤη */
    FixInst(l1, CurrentPC);

    /* 襹פѴ */
    ExitScope();
}

void Parse(void)
{
    GetToken();
    Prog();
    if (token != mc_eof)
	Error("ѿؿȤǧǤޤ");
}

int main (int argc, char *argv[])
{
    if (argc < 2) {
	fprintf(stderr, "ե뤬ꤵƤޤ\n");
	exit(1);
    }
    if (OpenFile(argv[1]) != 0) {
	fprintf(stderr, "եɤ߹ळȤǤޤ\n");
	exit(1);
    }
    Parse();
    CloseFile();
    if (WriteObjectFile("mcc.out") != 0) {
	fprintf(stderr, "֥ȥե뤳ȤǤޤ\n");
	exit(1);
    }
    return 0;
}

