/*
 *      parse1.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "scan.h"

#define MVM_R0  (0)
#define MVM_GP  (11)
#define MVM_FP  (13)

typedef unsigned long Word;
typedef long Offset;

typedef enum { Var, Fun, Unknown } ObjectType;

typedef struct Object {
    ObjectType      type;       /* ֥Ȥμ(ѿ/ؿ/̤) */
    char            name[MC_MAXIDLEN+1];        /* ѿ̾/ؿ̾ */
    int             reg;        /* ѿɥ쥹δȤʤ쥸ֹ
				   ؿξ R0ѿξ GP (R11)
				   ؿβɽѿξ FP (R13) */
    Offset          offset;     /* reg Υեå(а) */
    struct Object * next;       /* ƱפμΥ֥ */
} Object;

typedef struct Scope {
    struct Scope *  outer;      /* ĳ¦Υ */
    Object *        objects;    /* ΥפΥ֥ȤΥꥹ */
    Offset          offset;     /* ѿμγ() */
} Scope;

static Scope * GlobalScope;     /* (ֳ¦) */
static Scope * CurrentScope;    /* ߥѥΥ */

/* FindObject() : ΥפΥ֥ꥹȤ饪֥Ȥõ
		  Ф Ĥʤ 0 ֤ */

static Object * FindObject(Scope *scope, char *name)
{
    Object *obj;

    for (obj = scope->objects; obj != (Object *)0; obj = obj->next) {
	if (strcmp(name, obj->name) == 0)
	    return obj;
    }
    return (Object *)0;
}

/* NewObject() : ֥Ȥ롣 */

static Object * NewObject(ObjectType type, char *name)
{
    Object *obj;

    obj = (Object *) malloc(sizeof(Object));
    if (obj == (Object *)0)
	Error("ѥ뤹뤿Υ꤬­ޤ");
    strcpy(obj->name, name);
    obj->type = type;
    obj->next = (Object *)0;
    return obj;
}

/* NewVar() : ѿ򸽺ߤΥפɲä롣*/

static Object * NewVar(char *name)
{
    Object *obj;

    if (FindObject(CurrentScope, name) != (Object *)0)
	Error("ѿ̾ޤϰ̾ͤƤޤ");

    obj = NewObject(Var, name);
    obj->next = CurrentScope->objects;

    if (CurrentScope->outer == (Scope *)0) { /* ѿ */
	obj->reg = MVM_GP;
	obj->offset = CurrentScope->offset;
	CurrentScope->offset += sizeof(Word);
    }
    else if (CurrentScope->outer->outer == (Scope *)0) { /*  */
	obj->reg = MVM_FP;
	obj->offset = 0;        /*  FunDef() ꤹ */
    }
    else { /* ɽѿ */
	obj->reg = MVM_FP;
	CurrentScope->offset -= sizeof(Word);
	obj->offset = CurrentScope->offset;
    }

    CurrentScope->objects = obj;
    return obj;
}

/* UseObject() : ֥ȤõФ ĤʤС
		 type=Unknown Ȥ襹פ˥֥Ȥ롣 */

static Object * UseObject(char *name)
{
    Scope *scope;
    Object *obj;

    /* ¦Υפ襹פ˸äƽ˥֥Ȥõ */
    for (scope = CurrentScope; scope != (Scope *)0; scope = scope->outer) {
	obj = FindObject(scope, name);
	if (obj != (Object *)0)
		return obj;
    }

    /* 襹פ˿֥Ȥ */
    obj = NewObject(Unknown, name);
    obj->reg = MVM_R0;
    obj->offset = 0;
    obj->next = GlobalScope->objects;
    GlobalScope->objects = obj;
    return obj;
}

/* NewScope() : פ CurrentScope Ȥ롣 */

static void NewScope(Offset offset)
{
    Scope * scope;

    scope = (Scope *) malloc(sizeof(Scope));
    if (scope == (Scope *)0)
	Error("ѥ뤹뤿Υ꤬­ޤ");
    scope->outer = CurrentScope;
    scope->objects = (Object *)0;
    scope->offset = offset;
    CurrentScope = scope;
}

/* ExitScope() : CurrentScope Ѵĳ¦Υפ˽Ф롣 */

static void ExitScope(void)
{
    char message[100+MC_MAXIDLEN];
    Scope *cur = CurrentScope;
    Object *obj;

    /* ̤ؿΥå */
    for (obj = cur->objects; obj != (Object *)0; obj = obj->next) {
	if (obj->type == Unknown) {
	    sprintf(message, "ؿ %s ̤Ǥ", obj->name);
	    Error(message);
	}
    }

    CurrentScope = CurrentScope->outer;
    free((void *)cur);
}

static void Exp(void);
static void Block(void);

static void FunArgs(void)
{
    if (token != mc_lparen)
	Error("ؿƽФΰꥹȤ ( ޤ");
    GetToken();
    if (token != mc_rparen) {
	Exp();
	while (token == mc_comma) {
	    GetToken();
	    Exp();
	}
	if (token != mc_rparen)
	    Error("ؿƽФΰꥹȤ ) ޤ");
    }
    GetToken();
}

static void Factor(void)
{
    Object *obj;

    switch (token) {
    case mc_ident:
	/* ѿ: <Ident>ޤϡؿƤӽФ: <Ident> <FunArgs> */
	obj = UseObject(ident);
	GetToken();
	if (token == mc_lparen) {
	    if (obj->type == Var)
		Error("ѿФؿƽФϤǤޤ");
	    FunArgs();
	}
	else if (obj->type != Var)
	    Error("ƤʤѿϻѤǤޤ");
	break;
    case mc_number:
	/* : <Number> */
	GetToken();
	break;
    case mc_lparen:
	/* ̤ǰϤޤ줿: ( <Exp> ) */
	GetToken();
	Exp();
	if (token == mc_rparen)
	    GetToken();
	else
	    Error("Ĥ̤٤֤ˤޤ");
	break;
    default:
	Error("ʸˡ꤬ޤ");
	break;
    }
}

static void Term(void)
{
    Factor();
    while (token == mc_mult || token == mc_div || token == mc_mod) {
	GetToken();
	Factor();
    }
}

static void Exp(void)
{
    if (token == mc_plus || token == mc_minus) {
	GetToken();
	Term();
    }
    else
	Term();
    while (token == mc_plus || token == mc_minus) {
	GetToken();
	Term();
    }
}

static void Cond(void)
{
    Exp();
    if (mc_eq <= token && token <= mc_ge) {
	GetToken();
	Exp();
    }
    else
	Error("Ｐʸˡ꤬ޤ");
}

static void Statement(void)
{
    Object *obj;

    switch (token) {
    case mc_ident:
	/* <Ident> = <Exp>; | <Ident> <FunArgs>; */
	obj = UseObject(ident);
	GetToken();
	switch (token) {
	case mc_assign:
	    if (obj->type != Var)
		Error("ƤʤѿؤϤǤޤ");
	    GetToken();
	    Exp();
	    if (token != mc_semicolon)
		Error("ʸ ; ޤ");
	    GetToken();
	    break;
	case mc_lparen:
	    if (obj->type == Var)
		Error("ѿФؿƽФϤǤޤ");
	    FunArgs();
	    if (token != mc_semicolon)
		Error("ؿƽФʸ ; ޤ");
	    GetToken();
	    break;
	default:
	    Error("ʸ =ޤϡؿƽФʸ ( ޤ");
	    break;
	}
	break;
    case mc_return:
	/* return <Exp>; */
	GetToken();
	Exp();
	if (token != mc_semicolon)
	    Error("return ʸ ; ޤ");
	GetToken();
	break;
    case mc_input:
	/* input <Ident>; */
	GetToken();
	if (token != mc_ident)
	    Error("input ʸѿ̾ޤ");
	obj = UseObject(ident);
	GetToken();
	if (token != mc_semicolon)
	    Error("input ʸ ; ޤ");
	GetToken();
	break;
    case mc_print:
	/* print <Exp>; */
	GetToken();
	Exp();
	if (token != mc_semicolon)
	    Error("print ʸ ; ޤ");
	GetToken();
	break;
    case mc_if:
	/* if ( <Cond> ) <Statement> [ else <Statement> ] */
	GetToken();
	if (token != mc_lparen)
	    Error("if θ ( ޤ");
	GetToken();
	Cond();
	if (token != mc_rparen)
	    Error("if ξＰθ ) ޤ");
	GetToken();
	Statement();
	if (token == mc_else) {
	    GetToken();
	    Statement();
	}
	break;
    case mc_while:
	/* while ( <Cond> ) <Statement> */
	GetToken();
	if (token != mc_lparen)
	    Error("while θ ( ޤ");
	GetToken();
	Cond();
	if (token != mc_rparen)
	    Error("while ξＰθ ) ޤ");
	GetToken();
	Statement();
	break;
    case mc_lcb:
	/* "{" [ <VarDef> ] { <Statement> } "}" */
	Block();
	break;
    case mc_semicolon:
	Error("ʸϵƤޤ");
	break;
    default:
	Error("ʸϤ٤˸꤬ޤ");
	break;
    }
}

static void VarDef(void)
{
    GetToken();
    if (token != mc_ident)
	Error("ѿ̾ޤ");
    NewVar(ident);
    GetToken();
    while (token == mc_comma) {
	GetToken();
	if (token == mc_ident) {
	    NewVar(ident);
	    GetToken();
	}
    }
    if (token != mc_semicolon)
	Error("ѿ ,  ; ޤ");
    GetToken();
}

static void FunDef(void)
{
    Offset offset;
    Object *obj;

    obj = UseObject(ident);
    if (obj->type == Var)
	Error("ѿ̾ؿ̾Ȥƺ뤳ȤϤǤޤ");
    if (obj->type == Fun)
	Error("ؿ뤳ȤϤǤޤ");
    obj->type = Fun;
    GetToken();

    NewScope((Offset)0);

    if (token != mc_lparen)
	Error("ؿΰꥹȤ ( ޤ");
    GetToken();
    if (token != mc_rparen) {
	if (token != mc_ident)
	    Error("ؿΰꥹȤʸˡ꤬ޤ");
	obj = NewVar(ident);
	GetToken();
	while (token == mc_comma) {
	    GetToken();
	    if (token != mc_ident)
		Error("ؿΰꥹȤʸˡ꤬ޤ");
	    obj = NewVar(ident);
	    GetToken();
	}
	if (token != mc_rparen)
	    Error("ؿΰꥹȤ ) ޤ");
    }
    GetToken();

    offset = 2 * sizeof(Word);  /* FP ǸΰޤǤΥեå */
    /* ǸΰˡΥեåȤ */
    for (obj = CurrentScope->objects; obj != (Object *)0; obj = obj->next) {
	obj->offset = offset;
	offset += sizeof(Word);
    }

    Block();

    ExitScope();
}

static void Block(void)
{
    NewScope(CurrentScope->offset);

    GetToken();
    if (token == mc_int)
	VarDef();
    while (token != mc_rcb)
	Statement();
    GetToken();

    ExitScope();
}

static void Prog(void)
{
    CurrentScope = (Scope *)0;
    NewScope((Offset)0);
    GlobalScope = CurrentScope;

    while (token == mc_int || token == mc_ident) {
	if (token == mc_int)
	    VarDef();
	else
	    FunDef();
    }

    ExitScope();
}

void Parse(void)
{
    GetToken();
    Prog();
    if (token != mc_eof)
	Error("ѿؿȤǧǤޤ");
}

int main (int argc, char *argv[])
{
    if (argc < 2) {
	fprintf(stderr, "ե뤬ꤵƤޤ\n");
	exit(1);
    }
    if (OpenFile(argv[1]) != 0) {
	fprintf(stderr, "եɤ߹ळȤǤޤ\n");
	exit(1);
    }
    Parse();
    CloseFile();
    return 0;
}

