/*
 *      scan.c
 */

#include <stdio.h>
#include <string.h>
#include "scan.h"

Token           token;
char *          ident;
unsigned long   number;
unsigned int    line;

static FILE *   source;
static int      ch;

typedef struct {
    char *      string;
    Token       token;
} Keyword;

static Keyword keywords[] = {
    {   "int",          mc_int          },
    {   "return",       mc_return       },
    {   "input",        mc_input        },
    {   "print",        mc_print        },
    {   "while",        mc_while        },
    {   "if",           mc_if           },
    {   "else",         mc_else         },
    {   0,              mc_null         }
};

void Error(char *message)
{
    fprintf (stderr, "%d: %s\n", line, message);
    exit(1);
}

static void GetChar(void)
{
    if (ch == EOF)
	return;
    if (ch == '\n')
	line ++;
    ch = getc(source);
}

static void Identifier(void)
{
    static char buf[MC_MAXIDLEN+1];
    int i = 0;

    /* ̻Ҥ뤤ͽ buf ɤ߼ */
    do {
	if (MC_MAXIDLEN <= i)
	    Error("̻ҤĹޤ");
	buf[i++] = ch;
	GetChar();
    } while ('0' <= ch && ch <= '9' || ch == '_'
	     || 'A' <= ch && ch <= 'Z' || 'a' <= ch && ch <= 'z');
    buf[i] = '\0';

    /* ͽɽǸ */
    for (i = 0; keywords[i].string != (char *)0; i ++) {
	if (strcmp(keywords[i].string, buf) == 0) {
	    token = keywords[i].token;
	    return;
	}
    }

    /* ͽǤʤΤǼ̻ */
    token = mc_ident;
    ident = buf;
}

static void Number(void)
{
    unsigned long n = 0;
    int d;

    do {
	d = ch - '0';
	if ((MC_MAXNUM - d) / 10 < n)
	    Error("ͤ礭᤮ޤ");
	n = 10*n + d;
	GetChar();
    } while ('0' <= ch && ch <= '9');

    token = mc_number;
    number = n;
}

static void Comment(void)
{
    /* ȤɤߤȤФ */
    GetChar();
    do {
	while (ch != '*' && ch != EOF)
	    GetChar();
	GetChar();
    } while (ch != '/' && ch != EOF);
    GetChar();
}

void GetToken(void)
{
    /* ɤߤȤФ */
    while (ch <= ' ') {
	if (ch == EOF) {
	    token = mc_eof;
	    return;
	}
	GetChar();
    }

    /* ȡμ */
    switch (ch) {
	case '(':       GetChar(); token = mc_lparen; break;
	case ')':       GetChar(); token = mc_rparen; break;
	case '{':       GetChar(); token = mc_lcb; break;
	case '}':       GetChar(); token = mc_rcb; break;
	case '*':       GetChar(); token = mc_mult; break;
	case '%':       GetChar(); token = mc_mod; break;
	case ',':       GetChar(); token = mc_comma; break;
	case ';':       GetChar(); token = mc_semicolon; break;
	case '+':       GetChar();
                        if (ch == '+') {
                            GetChar();
                            token = mc_plusplus;
                        }
                        else
                            token = mc_plus;
                        break;
	case '-':       GetChar();
                        if (ch == '-') {
                            GetChar();
                            token = mc_minusminus;
                        }
                        else
                            token = mc_minus;
                        break;
	case '!':       GetChar();
			if (ch != '=')
			    Error("! μ = ǤʤФʤޤ");
			GetChar();
			token = mc_neq;
			break;
	case '=':       GetChar();
			if (ch == '=') {
			    GetChar();
			    token = mc_eq;
			}
			else
			    token = mc_assign;
			break;
	case '<':       GetChar();
			if (ch == '=') {
			    GetChar();
			    token = mc_le;
			}
			else
			    token = mc_lt;
			break;
	case '>':       GetChar();
			if (ch == '=') {
			    GetChar();
			    token = mc_ge;
			}
			else
			    token = mc_gt;
			break;
	case '/':       GetChar();
			if (ch == '*') {
			    Comment();
			    GetToken();
			}
			else
			    token = mc_div;
			break;
	case '0': case '1': case '2': case '3': case '4':
	case '5': case '6': case '7': case '8': case '9':
			Number();
			break;
	case 'A': case 'B': case 'C': case 'D': case 'E':
	case 'F': case 'G': case 'H': case 'I': case 'J':
	case 'K': case 'L': case 'M': case 'N': case 'O':
	case 'P': case 'Q': case 'R': case 'S': case 'T':
	case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
	case 'a': case 'b': case 'c': case 'd': case 'e':
	case 'f': case 'g': case 'h': case 'i': case 'j':
	case 'k': case 'l': case 'm': case 'n': case 'o':
	case 'p': case 'q': case 'r': case 's': case 't':
	case 'u': case 'v': case 'w': case 'x': case 'y': case 'z':
			Identifier();
			break;
	default:        Error("̵̣ʸޤ");
			break;
    }
}

int OpenFile(char * filename)
{
    source = fopen(filename, "r");
    if (source == NULL)
	return -1;
    line = 1;
    ch = ' ';
    return 0;
}

void CloseFile(void)
{
    fclose(source);
}

